﻿using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;


using servicehost.contract;

namespace myservice
{
    [Service]
    public class WordcountingController
    {
        [EntryPoint(HttpMethods.Get, "/api/v1/wordcount")]
        public int Count(string text) {
            var wc = new Wordcounting();
            return wc.Count(text);
        }
        
        
        public class CountRequest {
            public string Text;
            public string[] Stopwords;
        }
        
        public class CountResponse {
            public int Total;
            public int Unique;
        }

        
        [EntryPoint(HttpMethods.Get, "/api/v2/wordcount")]
        public CountResponse Count([Payload]CountRequest req) {
            var wc = new Wordcounting();
            var r = wc.Count(req.Text, req.Stopwords);
            __totalWordsCounted += r.total;
            return new CountResponse { Total = r.total, Unique = r.unique};
        }

        
        private static int __totalWordsCounted = 0;

        [EntryPoint(HttpMethods.Get, "/api/v2/stats")]
        public int Stats() {
            return __totalWordsCounted;
        }
        
        
        public class Response<T> {
            public bool Success = true;
            public T Result;

            public string FailureMessage;
        }

        [EntryPoint(HttpMethods.Get, "/math/divide")]
        public Response<int> Divide(int a, int b) {
            try {
                var result = Math.Divide(a, b);
                return new Response<int> {Result = result};
            }
            catch (Exception ex) {
                return new Response<int> {Success = false, FailureMessage = ex.Message};
            }
        }
    }

    public class Math
    {
        public static int Divide(int a, int b)
        {
            if (b == 0) 
                throw new InvalidOperationException("Division by zero not allowed!");
            
            return a / b;
        }
    }
    
    
    public class Wordcounting
    {
        public int Count(string text) {
            var (total, _) = Count(text, new string[0]);
            return total;
        }

        public (int total, int unique) Count(string text, string[] stopwords) {
            var allWords = Extract_words(text);
            var relevantWords = Subtract_stopwords(allWords, stopwords).ToArray();
            return (relevantWords.Length, 
                    relevantWords.Distinct().Count());
        }

        private static IEnumerable<string> Extract_words(string text) {
            var chunks = Split_into_word_candidates(text);
            return Discard_non_words(chunks);
        }
        
        private static IEnumerable<string> Split_into_word_candidates(string text)
            => text.Split(new[] {' ', '\n', '\r', '\t'}, StringSplitOptions.RemoveEmptyEntries);

        private static IEnumerable<string> Discard_non_words(IEnumerable<string> chunks)
            => chunks.Where(ch => char.IsLetter(ch[0])).ToArray();

        private static IEnumerable<string> Subtract_stopwords(IEnumerable<string> words, string[] stopwords)
            => words.Where(w => !stopwords.Contains(w));
    }
    
    
    [TestFixture]
    public class Wordcounting_tests
    {
        [Test]
        public void Count()
        {
            var sut = new Wordcounting();
            var result = sut.Count("a b2\n1c   d,\tefg");
            Assert.AreEqual(4, result);
        }

        [Test]
        public void Count_wo_stopwords()
        {
            var sut = new Wordcounting();
            var(n,m) = sut.Count("the quick brown fox jumps over the lazy dog", new[]{"brown", "lazy"});
            Assert.AreEqual(7, n);
            Assert.AreEqual(6, m);
        }
    }
}